//////////////////////////////////////////////////////////////////
// Symbolic
//
// Library for the construction of nested symbolic propositions
// The Flying Pig!
// Started 5/8/2003
// http://www.flypig.co.uk
// 
// 
// The MIT License
// See symbolic.h, COPYING file or website for licence
// 
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
// Includes

#include "symbolic.h"
#include "symbolic_private.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <math.h>

//////////////////////////////////////////////////////////////////
// Defines

//////////////////////////////////////////////////////////////////
// Structures

//////////////////////////////////////////////////////////////////
// Global variables

//////////////////////////////////////////////////////////////////
// Function prototypes

int         gcd                     (int nA, int nB);
Operation * SimplifyNEGATIVE (Operation * psOp, Operation * psVar1);
Operation * SimplifyNOT (Operation * psOp, Operation * psVar1);
Operation * SimplifyADD (Operation * psOp, Operation * psVar1, Operation * psVar2);
Operation * SimplifySUB (Operation * psOp, Operation * psVar1, Operation * psVar2);
Operation * SimplifyMUL (Operation * psOp, Operation * psVar1, Operation * psVar2);
Operation * SimplifyDIV (Operation * psOp, Operation * psVar1, Operation * psVar2);
Operation * SimplifyDIVIDE (Operation * psOp, Operation * psVar1, Operation * psVar2);
Operation * SimplifyPOW (Operation * psOp, Operation * psVar1, Operation * psVar2);

//////////////////////////////////////////////////////////////////
// Main application

//////////////////////////////////////////////////////////////////
// Recursively simplify an operation if possible
Operation * SimplifyOperation (Operation * psOp) {
  Operation         * psReturn = NULL;
  OpUnary           * psUna = NULL;
  OpBinary          * psBin = NULL;
  OpTernary         * psTer = NULL;
  Operation         * psTemp = NULL;

  if (psOp) {
    switch (psOp->eOpType) {
      case OPTYPE_MEMORY:
        psReturn = psOp;
        break;
      case OPTYPE_INTEGER:
        psReturn = psOp;
        break;
      case OPTYPE_TRUTHVALUE:
        psReturn = psOp;
        break;
      case OPTYPE_VARIABLE:
        psReturn = psOp;
        break;
      case OPTYPE_UNARY:
        psReturn = psOp;
        psUna = psReturn->Vars.psUnary;
        psUna->psVar1 = SimplifyOperation (psUna->psVar1);
        switch (psUna->eOpType) {
          case OPUNARY_NEGATIVE:
            psReturn = SimplifyNEGATIVE (psOp, psUna->psVar1);
            break;
          case OPUNARY_NOT:
            psReturn = SimplifyNOT (psOp, psUna->psVar1);
            break;
          case OPUNARY_LN:
            if ((psUna->psVar1->eOpType == OPTYPE_INTEGER) && (psUna->psVar1->Vars.nInteger == 1)) {
              psReturn = CreateInteger (0);
              FreeRecursive (psOp);
            }
            break;
          default:
            // Do nothing
            break;
        }
        break;
      case OPTYPE_BINARY:
        psBin = psOp->Vars.psBinary;
        psBin->psVar1 = SimplifyOperation (psBin->psVar1);
        psBin->psVar2 = SimplifyOperation (psBin->psVar2);
        psReturn = psOp;
        switch (psBin->eOpType) {
          case OPBINARY_LAND:
            if (((psBin->psVar1->eOpType == OPTYPE_TRUTHVALUE)
              && (!psBin->psVar1->Vars.boTruth))
              || ((psBin->psVar2->eOpType == OPTYPE_TRUTHVALUE)
              && (!psBin->psVar2->Vars.boTruth))) {
              psReturn = CreateTruthValue (FALSE);
              FreeRecursive (psOp);
            }
            else {
              if (((psBin->psVar1->eOpType == OPTYPE_TRUTHVALUE)
                && (psBin->psVar1->Vars.boTruth))
                || (CompareOperations (psBin->psVar1, psBin->psVar2))) {
                psReturn = CopyRecursive (psBin->psVar2);
                FreeRecursive (psOp);
              }
              else {
                if ((psBin->psVar2->eOpType == OPTYPE_TRUTHVALUE)
                  && (psBin->psVar2->Vars.boTruth)) {
                  psReturn = CopyRecursive (psBin->psVar1);
                  FreeRecursive (psOp);
                }
              }
            }
            break;
          case OPBINARY_LOR:
            if (((psBin->psVar1->eOpType == OPTYPE_TRUTHVALUE)
              && (psBin->psVar1->Vars.boTruth))
              || ((psBin->psVar2->eOpType == OPTYPE_TRUTHVALUE)
              && (psBin->psVar2->Vars.boTruth))) {
              psReturn = CreateTruthValue (TRUE);
              FreeRecursive (psOp);
            }
            else {
              if ((psBin->psVar1->eOpType == OPTYPE_TRUTHVALUE)
                && (!psBin->psVar1->Vars.boTruth)
                && (psBin->psVar2->eOpType == OPTYPE_TRUTHVALUE)
                && (!psBin->psVar2->Vars.boTruth)) {
                psReturn = CreateTruthValue (FALSE);
                FreeRecursive (psOp);
              }
            }
            break;
          case OPBINARY_LIMP:
            if ((((psBin->psVar1->eOpType == OPTYPE_TRUTHVALUE)
              && (!psBin->psVar1->Vars.boTruth))
              || ((psBin->psVar2->eOpType == OPTYPE_TRUTHVALUE)
              && (psBin->psVar2->Vars.boTruth)))
              || CompareOperations (psBin->psVar1, psBin->psVar2)) {
              psReturn = CreateTruthValue (TRUE);
              FreeRecursive (psOp);
            }
            else {
              if ((psBin->psVar1->eOpType == OPTYPE_TRUTHVALUE)
                && (psBin->psVar1->Vars.boTruth)) {
                psReturn = CopyRecursive (psBin->psVar2);
                FreeRecursive (psOp);
              }
            }
            break;
          case OPBINARY_POW:
            psReturn = SimplifyPOW (psOp, psBin->psVar1, psBin->psVar2);
            break;
          case OPBINARY_ADD:
            psReturn = SimplifyADD (psOp, psBin->psVar1, psBin->psVar2);
            break;
          case OPBINARY_SUB:
            psReturn = SimplifySUB (psOp, psBin->psVar1, psBin->psVar2);
            break;
          case OPBINARY_MUL:
            psReturn = SimplifyMUL (psOp, psBin->psVar1, psBin->psVar2);
            break;
          case OPBINARY_DIV:
            psReturn = SimplifyDIV (psOp, psBin->psVar1, psBin->psVar2);
            break;
          case OPBINARY_DIVIDE:
            psReturn = SimplifyDIVIDE (psOp, psBin->psVar1, psBin->psVar2);
            break;
          case OPBINARY_MOD:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              psReturn = CreateInteger (psBin->psVar1->Vars.nInteger
                % psBin->psVar2->Vars.nInteger);
              FreeRecursive (psOp);
            }
            break;
          case OPBINARY_OR:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              psReturn = CreateInteger (psBin->psVar1->Vars.nInteger
                | psBin->psVar2->Vars.nInteger);
              FreeRecursive (psOp);
            }
            break;
          case OPBINARY_AND:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              psReturn = CreateInteger (psBin->psVar1->Vars.nInteger
                & psBin->psVar2->Vars.nInteger);
              FreeRecursive (psOp);
            }
            break;
          case OPBINARY_BIC:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              psReturn = CreateInteger (psBin->psVar1->Vars.nInteger
                & !psBin->psVar2->Vars.nInteger);
              FreeRecursive (psOp);
            }
            break;
          case OPBINARY_EOR:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              psReturn = CreateInteger (psBin->psVar1->Vars.nInteger
                ^ psBin->psVar2->Vars.nInteger);
              FreeRecursive (psOp);
            }
            break;
          case OPBINARY_ROR:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              psReturn = CreateInteger (psBin->psVar1->Vars.nInteger
                >> psBin->psVar2->Vars.nInteger);
              FreeRecursive (psOp);
            }
            break;
          case OPBINARY_ROL:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              psReturn = CreateInteger (psBin->psVar1->Vars.nInteger
                << psBin->psVar2->Vars.nInteger);
              FreeRecursive (psOp);
            }
            break;
          case OPBINARY_EQ:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              if (psBin->psVar1->Vars.nInteger == psBin->psVar2->Vars.nInteger) {
                psReturn = CreateTruthValue (TRUE);
                FreeRecursive (psOp);
              }
              else {
                psReturn = CreateTruthValue (FALSE);
                FreeRecursive (psOp);
              }
            }
            else {
              if (CompareOperations (psBin->psVar1, psBin->psVar2)) {
                psReturn = CreateTruthValue (TRUE);
                FreeRecursive (psOp);
              }
            }
            break;
          case OPBINARY_NE:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              if (psBin->psVar1->Vars.nInteger
                != psBin->psVar2->Vars.nInteger) {
                psReturn = CreateTruthValue (TRUE);
                FreeRecursive (psOp);
              }
              else {
                psReturn = CreateTruthValue (FALSE);
                FreeRecursive (psOp);
              }
            }
            else {
              if (CompareOperations (psBin->psVar1, psBin->psVar2)) {
                psReturn = CreateTruthValue (FALSE);
                FreeRecursive (psOp);
              }
            }
            break;
          case OPBINARY_LT:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              if (psBin->psVar1->Vars.nInteger
                < psBin->psVar2->Vars.nInteger) {
                psReturn = CreateTruthValue (TRUE);
                FreeRecursive (psOp);
              }
              else {
                psReturn = CreateTruthValue (FALSE);
                FreeRecursive (psOp);
              }
            }
            break;
          case OPBINARY_LE:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              if (psBin->psVar1->Vars.nInteger
                <= psBin->psVar2->Vars.nInteger) {
                psReturn = CreateTruthValue (TRUE);
                FreeRecursive (psOp);
              }
              else {
                psReturn = CreateTruthValue (FALSE);
                FreeRecursive (psOp);
              }
            }
            break;
          case OPBINARY_GT:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              if (psBin->psVar1->Vars.nInteger
                > psBin->psVar2->Vars.nInteger) {
                psReturn = CreateTruthValue (TRUE);
                FreeRecursive (psOp);
              }
              else {
                psReturn = CreateTruthValue (FALSE);
                FreeRecursive (psOp);
              }
            }
            break;
          case OPBINARY_GE:
            if ((psBin->psVar1->eOpType == OPTYPE_INTEGER)
              && (psBin->psVar2->eOpType == OPTYPE_INTEGER)) {
              if (psBin->psVar1->Vars.nInteger
                >= psBin->psVar2->Vars.nInteger) {
                psReturn = CreateTruthValue (TRUE);
                FreeRecursive (psOp);
              }
              else {
                psReturn = CreateTruthValue (FALSE);
                FreeRecursive (psOp);
              }
            }
            break;
          case OPBINARY_IND:
            // Search for a 'set' reference in the memory
            psTemp = psBin->psVar1;
            while ((psTemp->eOpType == OPTYPE_TERNARY)
              && (psTemp->Vars.psTernary->eOpType == OPTERNARY_SET)
              && (!CompareOperations (psBin->psVar2,
              psTemp->Vars.psTernary->psVar2))
              && ((psTemp->Vars.psTernary->psVar2->eOpType != OPTYPE_BINARY)
              || ((psTemp->Vars.psTernary->psVar2->eOpType == OPTYPE_BINARY)
              && (psTemp->Vars.psTernary->psVar2->Vars.psBinary->eOpType
              != OPBINARY_IND)))) {
              psTemp = psTemp->Vars.psTernary->psVar1;
            }
            if ((psTemp->eOpType == OPTYPE_TERNARY)
              && (psTemp->Vars.psTernary->eOpType == OPTERNARY_SET)
              && (CompareOperations (psBin->psVar2,
              psTemp->Vars.psTernary->psVar2))
              && ((psTemp->Vars.psTernary->psVar2->eOpType != OPTYPE_BINARY)
              || ((psTemp->Vars.psTernary->psVar2->eOpType == OPTYPE_BINARY)
              && (psTemp->Vars.psTernary->psVar2->Vars.psBinary->eOpType
              != OPBINARY_IND)))) {
              psReturn = CopyRecursive (psTemp->Vars.psTernary->psVar3);
              FreeRecursive (psOp);
            }
//            psTemp = psBin->psVar1;
//            while ((psTemp->eOpType == OPTYPE_TERNARY)
//              && (psTemp->Vars.psTernary->eOpType == OPTERNARY_SET)
//              && (psTemp->Vars.psTernary->psVar2->eOpType == OPTYPE_BINARY)
//              && (psTemp->Vars.psTernary->psVar2->Vars.psBinary->eOpType
//              == OPBINARY_IND)
//              && (!CompareOperations (psBin->psVar2,
//              psTemp->Vars.psTernary->psVar2->Vars.psBinary->psVar2))
//              && ((psTemp->Vars.psTernary->psVar2->Vars.psBinary->psVar2->
//              eOpType != OPTYPE_BINARY)
//              || ((psTemp->Vars.psTernary->psVar2->Vars.psBinary->psVar2->
//              eOpType == OPTYPE_BINARY)
//              && (psTemp->Vars.psTernary->psVar2->Vars.psBinary->psVar2->
//              Vars.psBinary->eOpType != OPBINARY_IND))))
//            {
//              psTemp = psTemp->Vars.psTernary->psVar1;
//            }
//            if ((psTemp->eOpType == OPTYPE_TERNARY)
//              && (psTemp->Vars.psTernary->eOpType == OPTERNARY_SET)
//              && (psTemp->Vars.psTernary->psVar2->eOpType == OPTYPE_BINARY)
//              && (psTemp->Vars.psTernary->psVar2->Vars.psBinary->eOpType
//              == OPBINARY_IND)
//              && (CompareOperations (psBin->psVar2,
//              psTemp->Vars.psTernary->psVar2->Vars.psBinary->psVar2))
//              && ((psTemp->Vars.psTernary->psVar2->Vars.psBinary->psVar2->
//              eOpType != OPTYPE_BINARY)
//              || ((psTemp->Vars.psTernary->psVar2->Vars.psBinary->psVar2->
//              eOpType == OPTYPE_BINARY)
//              && (psTemp->Vars.psTernary->psVar2->Vars.psBinary->psVar2->
//              Vars.psBinary->eOpType != OPBINARY_IND))))
//            {
//              psReturn = CopyRecursive (psTemp->Vars.psTernary->psVar3);
//              FreeRecursive (psOp);
//            }
            break;
          default:
            break;
        }
        break;
      case OPTYPE_TERNARY:
        psTer = psOp->Vars.psTernary;
        psTer->psVar1 = SimplifyOperation (psTer->psVar1);

        psTer->psVar2 = SimplifyOperation (psTer->psVar2);
//        if ((psTer->eOpType == OPTERNARY_SET)
//          && (psTer->psVar2->eOpType == OPTYPE_BINARY)
//          && (psTer->psVar2->Vars.psBinary->eOpType == OPBINARY_IND))
//        {
//          // Skip the ind operation,
//          // since we can't have a set without an ind!
//          psTer->psVar2->Vars.psBinary->psVar1 =
//            SimplifyOperation (psTer->psVar2->Vars.psBinary->psVar1);
//          psTer->psVar2->Vars.psBinary->psVar2 =
//            SimplifyOperation (psTer->psVar2->Vars.psBinary->psVar2);
//        }
//        else
//        {
//          psTer->psVar2 = SimplifyOperation (psTer->psVar2);
//        }
        psTer->psVar3 = SimplifyOperation (psTer->psVar3);
        psReturn = psOp;
        switch (psTer->eOpType) {
          case OPTERNARY_SET:
            if ((psTer->psVar1->eOpType == OPTYPE_TERNARY)
              && (CompareOperations (psTer->psVar2,
              psTer->psVar1->Vars.psTernary->psVar2))) {
              psReturn = CreateTernary (OPTERNARY_SET,
                CopyRecursive (psTer->psVar1->Vars.psTernary->psVar1),
                CopyRecursive (psTer->psVar2),
                CopyRecursive (psTer->psVar3));
              FreeRecursive (psOp);
            }
            break;
          default:
            break;
        }
        break;
      default:
        psReturn = psOp;
        printf("Invalid operation type\n");
        break;
    }
  }
  return psReturn;
}

int gcd (int nA, int nB) {
  return (nB == 0 ? nA : gcd (nB, nA % nB));
}

Operation * SimplifyNEGATIVE (Operation * psOp, Operation * psVar1) {
  Operation * psReturn = psOp;
  OpBinary * psBin = NULL;

  if (psVar1->eOpType == OPTYPE_INTEGER) {
    psReturn = CreateInteger (- psVar1->Vars.nInteger);
    FreeRecursive (psOp);
  }
  else if ((psVar1->eOpType == OPTYPE_UNARY)
    && (psVar1->Vars.psUnary->eOpType == OPUNARY_NEGATIVE))  {
    // Remove pairs of 'negatives'
    psReturn = CopyRecursive (psVar1->Vars.psUnary->psVar1);
    FreeRecursive (psOp);
  }
  else if (psVar1->eOpType == OPTYPE_BINARY) {
    psBin = psVar1->Vars.psBinary;
    switch (psBin->eOpType) {
      case OPBINARY_SUB:
        // Switch round terms
        psReturn = CreateBinary (OPBINARY_SUB, CopyRecursive (psBin->psVar2),
          CopyRecursive (psBin->psVar1));
        FreeRecursive (psOp);
        break;
      default:
        // Do nothing
        break;
    }
  }

  return psReturn;
}

Operation * SimplifyNOT (Operation * psOp, Operation * psVar1) {
  Operation * psReturn = psOp;
  OpBinary * psBin = NULL;

  switch (psVar1->eOpType) {
    case OPTYPE_UNARY:
      if (psVar1->Vars.psUnary->eOpType == OPUNARY_NOT) {
        // Remove pairs of 'nots'
        psReturn = CopyRecursive (psVar1->Vars.psUnary->psVar1);
        FreeRecursive (psOp);
      }
      break;
    case OPTYPE_TRUTHVALUE:
      // Negate the truth value
      psReturn = CreateTruthValue (!psVar1->Vars.boTruth);
      FreeRecursive (psOp);
      break;
    case OPTYPE_BINARY:
      psBin = psVar1->Vars.psBinary;
      switch (psBin->eOpType) {
        case OPBINARY_EQ:
          // Change to NE
          psReturn = CreateBinary (OPBINARY_NE, psBin->psVar1, psBin->psVar2);
          FreeRecursive (psOp);
          break;
        case OPBINARY_NE:
          // Change to EQ
          psReturn = CreateBinary (OPBINARY_EQ, psBin->psVar1, psBin->psVar2);
          FreeRecursive (psOp);
          break;
        case OPBINARY_LT:
          // Change to GE
          psReturn = CreateBinary (OPBINARY_GE, psBin->psVar1, psBin->psVar2);
          FreeRecursive (psOp);
          break;
        case OPBINARY_LE:
          // Change to GT
          psReturn = CreateBinary (OPBINARY_GT, psBin->psVar1, psBin->psVar2);
          FreeRecursive (psOp);
          break;
        case OPBINARY_GT:
          // Change to LE
          psReturn = CreateBinary (OPBINARY_LE, psBin->psVar1,
            psBin->psVar2);
          FreeRecursive (psOp);
          break;
        case OPBINARY_GE:
          // Change to LT
          psReturn = CreateBinary (OPBINARY_LT, psBin->psVar1, psBin->psVar2);
          FreeRecursive (psOp);
          break;
        default:
          // Do nothing
          break;
      }
      break;
    default:
      // Do nothing
      break;
  }

  return psReturn;
}

Operation * SimplifyADD (Operation * psOp, Operation * psVar1, Operation * psVar2) {
  Operation * psReturn = psOp;

  if ((psVar1->eOpType == OPTYPE_INTEGER)
    && (psVar2->eOpType == OPTYPE_INTEGER)) {
    psReturn = CreateInteger (psVar1->Vars.nInteger
      + psVar2->Vars.nInteger);
    FreeRecursive (psOp);
  }
  else if ((psVar1->eOpType == OPTYPE_INTEGER)
    && (psVar1->Vars.nInteger == 0)) {
    psReturn = CopyRecursive (psVar2);
    FreeRecursive (psOp);
  }
  else if ((psVar2->eOpType == OPTYPE_INTEGER)
    && (psVar2->Vars.nInteger == 0)) {
    psReturn = CopyRecursive (psVar1);
    FreeRecursive (psOp);
  }
  else if ((psVar1->eOpType == OPTYPE_BINARY)
    && (psVar2->eOpType == OPTYPE_INTEGER)) {
    if ((psVar1->Vars.psBinary->eOpType == OPBINARY_ADD)
      && (psVar1->Vars.psBinary->psVar2->eOpType == OPTYPE_INTEGER)) {
      psReturn = CreateBinary (OPBINARY_ADD,
        CopyRecursive (psVar1->Vars.psBinary->psVar1),
        CreateInteger (psVar1->Vars.psBinary->psVar2->
        Vars.nInteger + psVar2->Vars.nInteger));
      FreeRecursive (psOp);
    }
    else if ((psVar1->Vars.psBinary->eOpType == OPBINARY_SUB)
      && (psVar1->Vars.psBinary->psVar2->eOpType == OPTYPE_INTEGER)) {
      if (psVar1->Vars.psBinary->psVar2->Vars.nInteger
        < psVar2->Vars.nInteger) {
        psReturn = CreateBinary (OPBINARY_ADD,
          CopyRecursive (psVar1->Vars.psBinary->psVar1),
          CreateInteger (psVar2->Vars.nInteger -
          psVar1->Vars.psBinary->psVar2->Vars.nInteger));
      }
      else {
        psReturn = CreateBinary (OPBINARY_SUB,
          CopyRecursive (psVar1->Vars.psBinary->psVar1),
          CreateInteger (psVar1->Vars.psBinary->psVar2->
          Vars.nInteger - psVar2->Vars.nInteger));
      }
      FreeRecursive (psOp);
    }
    else if ((psVar1->Vars.psBinary->eOpType == OPBINARY_DIVIDE)
      && (psVar1->Vars.psBinary->psVar1->eOpType == OPTYPE_INTEGER)
      && (psVar1->Vars.psBinary->psVar2->eOpType == OPTYPE_INTEGER)) {
      // Useful for simplifying continues fractions
      psReturn = CreateBinary (OPBINARY_DIVIDE, 
        CreateInteger (psVar1->Vars.psBinary->psVar1->Vars.nInteger
        + (psVar1->Vars.psBinary->psVar2->Vars.nInteger * psVar2->Vars.nInteger)),
        CopyRecursive (psVar1->Vars.psBinary->psVar2));
      FreeRecursive (psOp);
    }
  }
  else if ((psVar1->eOpType == OPTYPE_INTEGER)
    && (psVar2->eOpType == OPTYPE_BINARY)) {
    if ((psVar2->Vars.psBinary->eOpType == OPBINARY_ADD)
      && (psVar2->Vars.psBinary->psVar2->eOpType == OPTYPE_INTEGER)) {
      psReturn = CreateBinary (OPBINARY_ADD,
        CopyRecursive (psVar2->Vars.psBinary->psVar1),
        CreateInteger (psVar2->Vars.psBinary->psVar2->
        Vars.nInteger + psVar1->Vars.nInteger));
      FreeRecursive (psOp);
    }
    else if ((psVar2->Vars.psBinary->eOpType == OPBINARY_SUB)
      && (psVar2->Vars.psBinary->psVar2->eOpType == OPTYPE_INTEGER)) {
      if (psVar2->Vars.psBinary->psVar2->Vars.nInteger
        < psVar1->Vars.nInteger) {
        psReturn = CreateBinary (OPBINARY_ADD,
          CopyRecursive (psVar2->Vars.psBinary->psVar1),
          CreateInteger (psVar1->Vars.nInteger -
          psVar2->Vars.psBinary->psVar2->Vars.nInteger));
      }
      else {
        psReturn = CreateBinary (OPBINARY_SUB,
          CopyRecursive (psVar2->Vars.psBinary->psVar1),
          CreateInteger (psVar2->Vars.psBinary->psVar2->
          Vars.nInteger - psVar1->Vars.nInteger));
      }
      FreeRecursive (psOp);
    }
    else if ((psVar2->Vars.psBinary->eOpType == OPBINARY_DIVIDE)
      && (psVar2->Vars.psBinary->psVar1->eOpType == OPTYPE_INTEGER)
      && (psVar2->Vars.psBinary->psVar2->eOpType == OPTYPE_INTEGER)) {
      // Useful for simplifying continues fractions
      psReturn = CreateBinary (OPBINARY_DIVIDE, 
        CreateInteger (psVar2->Vars.psBinary->psVar1->Vars.nInteger
        + (psVar2->Vars.psBinary->psVar2->Vars.nInteger * psVar1->Vars.nInteger)),
        CopyRecursive (psVar2->Vars.psBinary->psVar2));
      FreeRecursive (psOp);
    }
  }
  else if ((psVar1->eOpType == OPTYPE_BINARY)
    && (psVar2->eOpType == OPTYPE_BINARY)
    && (psVar1->Vars.psBinary->eOpType == OPBINARY_DIVIDE)
    && (psVar2->Vars.psBinary->eOpType == OPBINARY_DIVIDE)
    && (CompareOperations (psVar1->Vars.psBinary->psVar2, psVar2->Vars.psBinary->psVar2))) {
    psReturn = CreateBinary (OPBINARY_DIVIDE, CreateBinary (OPBINARY_ADD,
      CopyRecursive (psVar1->Vars.psBinary->psVar1),
      CopyRecursive (psVar2->Vars.psBinary->psVar1)), CopyRecursive (psVar1->Vars.psBinary->psVar2));
    FreeRecursive (psOp);
  }
  else if ((psVar2->eOpType == OPTYPE_UNARY)
    && (psVar2->Vars.psUnary->eOpType == OPUNARY_NEGATIVE)) {
    // Change the addition of a negative in the second term into a subtraction
    psReturn = CreateBinary (OPBINARY_SUB,
      CopyRecursive (psVar1),
      CopyRecursive (psVar2->Vars.psUnary->psVar1));
    FreeRecursive (psOp);
  }
  else if ((psVar1->eOpType == OPTYPE_UNARY)
    && (psVar1->Vars.psUnary->eOpType == OPUNARY_NEGATIVE)) {
    // Change the addition of a negative in the first term into a subtraction with
    // the terms switched around
    psReturn = CreateBinary (OPBINARY_SUB,
      CopyRecursive (psVar2->Vars.psUnary->psVar1),
      CopyRecursive (psVar1));
    FreeRecursive (psOp);
  }

  return psReturn;
}

Operation * SimplifySUB (Operation * psOp, Operation * psVar1, Operation * psVar2) {
  Operation * psReturn = psOp;

  if ((psVar1->eOpType == OPTYPE_INTEGER) && (psVar2->eOpType == OPTYPE_INTEGER)) {
    psReturn = CreateInteger (psVar1->Vars.nInteger - psVar2->Vars.nInteger);
    FreeRecursive (psOp);
  }
  else if ((psVar2->eOpType == OPTYPE_UNARY)
    && (psVar2->Vars.psUnary->eOpType == OPUNARY_NEGATIVE)) {
    // Change the subtraction of a negitive in the second term into an addition
    psReturn = CreateBinary (OPBINARY_ADD,
      CopyRecursive (psVar1),
      CopyRecursive (psVar2->Vars.psUnary->psVar1));
    FreeRecursive (psOp);
  }

  return psReturn;
}

Operation * SimplifyMUL (Operation * psOp, Operation * psVar1, Operation * psVar2) {
  Operation * psReturn = psOp;

  if ((psVar1->eOpType == OPTYPE_INTEGER) && (psVar2->eOpType == OPTYPE_INTEGER)) {
    psReturn = CreateInteger (psVar1->Vars.nInteger * psVar2->Vars.nInteger);
    FreeRecursive (psOp);
  }
  else if (psVar1->eOpType == OPTYPE_INTEGER) {
    if (psVar1->Vars.nInteger == 0) {
      psReturn = CreateInteger (0);
      FreeRecursive (psOp);
    }
    else if (psVar1->Vars.nInteger == 1) {
      psReturn = CopyRecursive (psVar2);
      FreeRecursive (psOp);
    }
  }
  else if (psVar2->eOpType == OPTYPE_INTEGER) {
    if (psVar2->Vars.nInteger == 0) {
      psReturn = CreateInteger (0);
      FreeRecursive (psOp);
    }
    else if (psVar2->Vars.nInteger == 1) {
      psReturn = CopyRecursive (psVar1);
      FreeRecursive (psOp);
    }
  }

  return psReturn;
}

Operation * SimplifyDIV (Operation * psOp, Operation * psVar1, Operation * psVar2) {
  Operation * psReturn = psOp;

  if ((psVar2->eOpType == OPTYPE_INTEGER) && (psVar2->Vars.nInteger == 1)) {
    psReturn = CopyRecursive (psVar1);
    FreeRecursive (psOp);
  }
  else if ((psVar1->eOpType == OPTYPE_INTEGER) && (psVar2->eOpType == OPTYPE_INTEGER)) {
    psReturn = CreateInteger (psVar1->Vars.nInteger / psVar2->Vars.nInteger);
    FreeRecursive (psOp);
  }

  return psReturn;
}

Operation * SimplifyDIVIDE (Operation * psOp, Operation * psVar1, Operation * psVar2) {
  Operation * psReturn = psOp;

  if ((psVar2->eOpType == OPTYPE_INTEGER)
    && (psVar2->Vars.nInteger == 1)) {
    psReturn = CopyRecursive (psVar1);
    FreeRecursive (psOp);
  }
  else if ((psVar1->eOpType == OPTYPE_INTEGER)
    && (psVar2->eOpType == OPTYPE_INTEGER)) {
    // Reduce the division if possible
    int nGcd;
    nGcd = gcd (psVar1->Vars.nInteger, psVar2->Vars.nInteger);
    if (nGcd == psVar2->Vars.nInteger) {
      // The denominator divides the numerator precisely
      psReturn = CreateInteger (psVar1->Vars.nInteger
        / psVar2->Vars.nInteger);
      FreeRecursive (psOp);
    }
    else {
      psVar1->Vars.nInteger /= nGcd;
      psVar2->Vars.nInteger /= nGcd;
    }
  }
  else if ((psVar2->eOpType == OPTYPE_BINARY)
    && (psVar2->Vars.psBinary->eOpType == OPBINARY_DIVIDE)) {
    psReturn = CreateBinary (OPBINARY_DIVIDE, 
      CreateBinary (OPBINARY_MUL, CopyRecursive (psVar1), 
      CopyRecursive (psVar2->Vars.psBinary->psVar2)), 
      CopyRecursive (psVar2->Vars.psBinary->psVar1));
    FreeRecursive (psOp);
  }
  else if ((psVar1->eOpType == OPTYPE_BINARY)
    && (psVar1->Vars.psBinary->eOpType == OPBINARY_DIVIDE)) {
    psReturn = CreateBinary (OPBINARY_DIVIDE, 
      CopyRecursive (psVar1->Vars.psBinary->psVar1),
      CreateBinary (OPBINARY_MUL, CopyRecursive (psVar2), 
      CopyRecursive (psVar1->Vars.psBinary->psVar2))); 
    FreeRecursive (psOp);
  }

  return psReturn;
}

Operation * SimplifyPOW (Operation * psOp, Operation * psVar1, Operation * psVar2) {
  Operation * psReturn = psOp;

  if (psVar2->eOpType == OPTYPE_INTEGER) {
    if (psVar2->Vars.nInteger == 0) {
      psReturn = CreateInteger (1);
      FreeRecursive (psOp);
    }
    else if (psVar2->Vars.nInteger == 1) {
      psReturn = CopyRecursive (psVar1);
      FreeRecursive (psOp);
    }
    else if (psVar1->eOpType == OPTYPE_INTEGER) {
      if (psVar2->Vars.nInteger > 0) {
        psReturn = CreateInteger ((int)pow ((double)psVar1->Vars.nInteger,
          (double)psVar2->Vars.nInteger));
        FreeRecursive (psOp);
      }
      else {
        psReturn = CreateBinary (OPBINARY_DIVIDE, CreateInteger (1), 
          CreateInteger ((int)pow ((double)psVar1->Vars.nInteger,
          (double)(-psVar2->Vars.nInteger))));
        FreeRecursive (psOp);
      }
    }
  }

  return psReturn;
}


Operation * UberSimplifyOperation (Operation * psOp) {
  Operation * psPrev = NULL;

  do {
    FreeRecursive (psPrev);
    psPrev = CopyRecursive (psOp);
    psOp = SimplifyOperation (psOp);
  } while (!CompareOperations (psOp, psPrev));

  FreeRecursive (psPrev);

  return psOp;
}
